(* ---------------------------------------------------------------------------- *)

(* An atom is an object with unique identity. *)

(* Atoms are heap-allocated. This gives us a number of good properties: the
   number of atoms is limited only by available memory; unused atoms are
   collected; atoms can contain informational fields (for pretty-printing) as
   well as transient mutable fields (for copying). *)

(* In order to support efficient persistent maps over atoms, we still need
   each atom to carry a unique number. With 31 bit integers, we can produce
   two billion atoms. 64 bit integers could be used if necessary. *)

type atom = {

  (* The identifier that originally gave rise to this atom. *)

  identifier: Identifier.t;

  (* This atom's identity. *)

  number: int;

}

(* ---------------------------------------------------------------------------- *)

(* A global counter is used to produce unique numbers. *)

let counter =
  ref 0

let number () : int =
  let number = !counter in
  let successor = number + 1 in
  counter := successor;
  assert (successor <> 0); (*  *)
  number

(* ---------------------------------------------------------------------------- *)

(* [fresh] produces a fresh atom. *)

let fresh identifier =
  let rec a = {
    identifier = identifier;
    number = number();
  } in
  a

(* ---------------------------------------------------------------------------- *)

(* Accessors. *)

let identifier { identifier = identifier } = identifier

(* ---------------------------------------------------------------------------- *)

(* Comparison. *)

let equal =
  (==)

let compare a1 a2 =
  Pervasives.compare a1.number a2.number

(* ---------------------------------------------------------------------------- *)

(* Sets and maps over atoms. *)

module Atom = struct
  type t = atom
  let compare = compare
end

module AtomSet =
  Set.Make (Atom)

module AtomMap = struct

  include Map.Make (Atom)

  let to_list m =
    List.rev
      (fold
         (fun key data accu ->
	   (key, data) :: accu
         ) m []
      )
  let of_list m =
    List.fold_left
      (fun fields (key, data) ->
        add key data fields
      ) empty m
      
end

